#include <stdio.h>
#include <iostream.h>
#include <fcntl.h>
#include <sys/iofunc.h>
#include <sys/syspage.h>
#include "agefun.h"
#include "comand.h"
#include "coder.h"
//---------------------------------------------------------------------------------------------------------------------------
// буфер исходного файла, подлежащего декодированию
static char *sSource = NULL; 
// длина этого файла
static long nLength = 0;
// длина ключа декодирования
static int nKeyLen = 0;
// указатели начального и конечного ключей
key *bKey = NULL, *fKey = NULL;
//---------------------------------------------------------------------------------------------------------------------------
// Управление открым каналом (devctl):
int devctl( resmgr_context_t *ctp, io_devctl_t *msg, iofunc_ocb_t *ocb ) { 
    int sts;
    if( ( sts = iofunc_devctl_default( ctp, msg, ocb ) ) != _RESMGR_DEFAULT ) return sts;
    char *data = (char*)_DEVCTL_DATA( msg->i );
    int nbytes = 0;
    switch( msg->i.dcmd ) {
        // команда установки файла-источника
        case DCMD_CLUST_STS: {
            if( sSource != NULL ) delete sSource;
            int fd;
            if( ( fd = open( data, O_RDONLY ) ) < 0 ) return errno;
            // определяется длина исходного файла и он целиком считывается в буфер 
            struct stat buf;
            if( fstat( fd, &buf ) == -1 ) return errno;
            nLength = buf.st_size;
            if( ( sSource = new char[ nLength ] ) == NULL ) return ENOMEM;
            long k = read( fd, sSource, nLength );
            if( k < 0 ) return errno;
            if( k != nLength ) return EIO;
            close( fd );
            break;
        };           
        // команда установки длины ключа
        case DCMD_CLUST_STK: 
            if( ( nKeyLen = atoi( data ) ) == 0 ) return EINVAL;
            break;
        // команда запроса частоты процессора
        case DCMD_CLUST_GTF:
            sprintf( data, "%d", SYSPAGE_ENTRY( qtime )->cycles_per_sec );
            nbytes = strlen( data ) + 1;
            break;
        // любая другая команда - это ошибка
        default: return ENOSYS;
    };
    memset( &msg->o, 0, sizeof( msg->o ) );
    msg->o.nbytes = nbytes;
    return _RESMGR_PTR( ctp, &msg->o, sizeof( msg->o ) + nbytes  );
};
//---------------------------------------------------------------------------------------------------------------------------
// Запись параметров (ключей: начало-конец) в открытый канал
int write( resmgr_context_t *ctp, io_write_t *msg, iofunc_ocb_t *ocb ) {
    int status = iofunc_write_verify( ctp, msg, ocb, NULL ); 
    if( status != EOK ) return status;   
    if( msg->i.xtype & _IO_XTYPE_MASK != _IO_XTYPE_NONE ) return ENOSYS; 
    if( msg->i.nbytes != 2 * nKeyLen ) return EILSEQ;
    _Uint8t *buf = new _Uint8t[ 2 * nKeyLen ], *bk = buf, *fk = buf + nKeyLen;
    resmgr_msgread( ctp, buf, msg->i.nbytes, sizeof( msg->i ) );
    if( bKey != NULL ) delete bKey;
    if( ( bKey = new key( nKeyLen, bk ) ) == NULL ) return ENOMEM;
    if( fKey != NULL ) delete fKey;
    if( ( fKey = new key( nKeyLen, fk ) ) == NULL ) return ENOMEM; 
    delete buf;
    ocb->attr->flags |= IOFUNC_ATTR_MTIME | IOFUNC_ATTR_DIRTY_TIME;
    return  EOK;        
};    
//---------------------------------------------------------------------------------------------------------------------------
// Чтение открытого канала (результата)
int read( resmgr_context_t *ctp, io_read_t *msg, iofunc_ocb_t *ocb ) {
    int status = iofunc_read_verify( ctp, msg, ocb, NULL );
    if( status != EOK ) return status;
    if( msg->i.xtype & _IO_XTYPE_MASK != _IO_XTYPE_NONE ) return ENOSYS;
    if( sSource == NULL ||  nLength <= 0 || nKeyLen == 0 ||
        bKey == NULL || fKey == NULL || *bKey > *fKey ) return EDOM; 
    _Uint8t *buf = new _Uint8t[ MAX_RESULT ];
    if( buf == 0 ) return ENOMEM;
    int n = 0;
    char *sOut;
    while( *bKey != *fKey ) {
        // кодируется входная последовательность новым ключём
        if( ( sOut = bKey->code( sSource, nLength ) ) == NULL ) return ENOMEM;
        // проверяется на принадлежность строки "печатным" символам
        if( test( sOut, nLength ) ) {
            memcpy( buf + n, (_Uint8t*)*bKey, nKeyLen );
            n += nKeyLen;
        }; 
        delete sOut;
        // смещается ключ
        bKey->next();
    };
    MsgReply( ctp->rcvid, n, buf, n );
    ocb->attr->flags |= IOFUNC_ATTR_ATIME | IOFUNC_ATTR_DIRTY_TIME;
    delete buf; 
    return  _RESMGR_NOREPLY;
}; 
//---------------------------------------------------------------------------------------------------------------------------


