#if !defined( __CODER_H )
#define __CODER_H
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/neutrino.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <iostream.h>
//---------------------------------------------------------------------------------------------------------
#define VERS "vers.1.12"
const int MAX_RESULT = 1024;
const int KEY_OUT_INLINE = 5;
//---------------------------------------------------------------------------------------------------------
// класс ключа - байтовая последовательность
class key {
    int           k;
    _Uint8t* p;
public:
    key( int, _Uint8t *f = NULL );
    // чтение и загрузка ключа из файла
    key( char* );
    key( key& );
    ~key( void ) { if( p != NULL ) delete p; };
    operator _Uint8t*() { return p; };
    // присвоение ключей
    key& operator=( const key& k ); 
    // сдвиг ключа в сторону увеличения
    key& rshift( unsigned long long );
    inline key& rshift( int n ) { return rshift( (unsigned long long)n ); };
    // формирование "следующего" ключа при переборе
    inline key& next( void ) { return rshift( (unsigned long long)1 ); };
    // кодирование-декодирование строки
    char *code( char*, unsigned long );
    // сравнение двух ключей
    friend bool operator==( const key&, const key& );
    friend inline bool operator!=( const key& k1, const key& k2 ) { return !( k1 == k2 ); };
    friend bool operator<( const key& k1, const key& k2 ) { return ( memcmp( k1.p, k2.p, k1.k ) < 0 ); };
    friend bool operator>( const key& k1, const key& k2 ) { return ( memcmp( k1.p, k2.p, k1.k ) > 0 ); };
    // вывод ключа
    friend ostream& operator<<( ostream& s, key& k );
};
key::key( int len, _Uint8t *f ) {
    k = len; 
    p = new _Uint8t[ k ];
    _Uint8t *t = p;
    if( f == NULL ) memset( p, 0, k );
    else memcpy( p, f, k );
 };
key::key( key& o ) {
    k = o.k;
    p = new _Uint8t[ k ];
    _Uint8t *d = p, *s = o.p;
    for( int i = 0; i < k; s++, d++, i++ ) d = s;
};
key::key( char *fn ) {
    int fd = open( fn, O_RDONLY );
    if( fd < 0 ) perror( "open error" ), exit( EXIT_FAILURE );
    struct stat buf;
    if( fstat( fd , &buf ) < 0 ) perror( "fstat error" ), exit( EXIT_FAILURE );
    k = buf.st_size;
    if( read( fd, p, k ) != k ) perror( "read error" ), exit( EXIT_FAILURE );
    close( fd );
};
key& key::operator=( const key& o ) {
    if( this == &o ) return *this;
    if( p != NULL ) delete p;
    p = new _Uint8t[ k = o.k ];
    _Uint8t *d = p, *s = o.p;
    for( int i = 0; i < k; s++, d++, i++ ) d = s;
    return *this; 
};
key& key::rshift( unsigned long long shf ) {
   unsigned long long prev = shf;
   for( int j = k - 1; j >= 0; j-- ) {
      unsigned long long val = p[ j ] + prev;
      if( val <= 255 ) {
         p[ j ] = val;
         break; 
      }
      else {
         p[ j ] = (_Uint8t)( val & 0xFF );
         prev = val >> 8; 
      };
   };
   return *this;
};
char* key::code( char *s, unsigned long n ) {
    char *r = new char [ n ];
    if( r == NULL ) return NULL;
    for( unsigned long i = 0; i < n; i++ ) r[ i ] = (char)( (_Uint8t)s[ i ] ^ *( p +  i % k ) );     
    return r;
};
bool operator==( const key& k1, const key& k2 ) {
    if( k1.k != k2.k ) return false;
    _Uint8t *s = k1.p, *d = k2.p;
    for( int i = 0; i < k1.k; i++, s++, d++ ) if( *s != *d ) return false;
    return true;
};
ostream& operator<<( ostream& s, key& k ) {
   s << "<";
   for( int i = 0; i < k.k; i++ ) {
      s << hex << (int)( k.p[ i ] );
      if( i != k.k - 1 ) s << ",";
   };
   return s << ">" << dec << flush;
};
//---------------------------------------------------------------------------------------------------------
// проверка строки на принадлежность "печатным" символам
bool test( const char src[], unsigned long srclen ) {
   const char *p = src;
   for( unsigned long i = 0; i < srclen; i++, p++ ) { 
       if( *p == '\n' || *p == '\t' || *p == '\r' || ( *p >= ' ' && *p <= '~' ) ) continue;
       return false;
   };
   return true;
};
//---------------------------------------------------------------------------------------------------------
#endif
