#include <dirent.h>
#include <limits.h>
#include <sys/netmgr.h>
#include <math.h>
#include <pthread.h>
#include <sync.h>
#include "comand.h"
#include "coder.h"
//------------------------------------------------------------------------------------------------------------
// класс, представляющий отдельный хост кластера на мастере
class Sutelite {
friend class SutList;
    Sutelite *next;
public:
    char *name;						// имя хоста
    int fd;								// дескриптор хоста
    unsigned long long cps;		// скорость процессора хоста
    _Uint8t *buf;						// буфер приёмо-передачи обмена с хостом
    int          nbuf;					// длина данных в этом буфере
    Sutelite( char *host ) { name = strdup( host ); buf = NULL; next = NULL; fd = -1; };
    ~Sutelite( void ) { delete name; close( fd ); };
    char *host( void ) { return name; };
};
//------------------------------------------------------------------------------------------------------------
// класс - список текущих активных хостов кластера
class SutList {
    Sutelite  *begin;
public:
    SutList( void ) { begin = NULL; };
    ~SutList( void );
    Sutelite *head( void ) { return begin; };   
    Sutelite *next( Sutelite *p ) { return p->next; };
    void add( Sutelite* );
    int size( void );
};
void SutList::add( Sutelite *host ) {
    if( begin == NULL ) begin = host;
    else {
        Sutelite *t = begin;
        for( ; t->next != NULL; t = t->next );
        t->next = host;
    };
};
SutList::~SutList( void ) {
    while( begin != NULL ) { Sutelite *t = begin; delete begin; begin = t->next; }
};
int SutList::size( void ) {
    int n = 0;
    if( begin != NULL ) for(  Sutelite *t = begin; t != NULL; t = t->next, n++ );
    return n;
};
//------------------------------------------------------------------------------------------------------------
// список текущих активных хостов кластера, используемых задачей 
static SutList *hlist = new SutList;
// длина ключа - используется глобально только для выходной печати
static int keylen;
static pthread_barrier_t barrier;
//------------------------------------------------------------------------------------------------------------
void *exec( void *p ) {
    Sutelite *h = (Sutelite*)p;
    int rc;
    // запись диаппазона (начало-конец) для хоста - начало работы
    if( ( rc = write( h->fd, h->buf, 2 * keylen ) ) < 0 ) cout << h->name, perror( "write error!" ); 
    delete( h->buf );
    if( rc >= 0 ) {
        // создаётся временный буфер для приёма результата
        _Uint8t *rbuf = new _Uint8t[ MAX_RESULT ]; 
        if( ( h->nbuf = ( read( h->fd, rbuf, MAX_RESULT ) ) ) < 0 ) cout << h->name, perror( "read error" );
        h->buf = new _Uint8t[ h->nbuf ];
        memcpy( h->buf, rbuf, h->nbuf );
        delete rbuf;
    };
    pthread_barrier_wait( &barrier );
};
//------------------------------------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {
    cout << "Multi Cluster Decoder, " << VERS << endl;
    if( argc != 3 ) cout << "wrong arguments number." << endl, exit( EXIT_FAILURE );
    // 1-й параметр - файл для декодирования, 2-й параметр - длина ключа
    int fds = open( argv[ 1 ], O_RDONLY );
    if( fds < 0 ) cout << "input file not found: " << argv[ 1 ] << endl, exit( EXIT_FAILURE );
    close( fds );
    keylen = atoi( argv[ 2 ] );
    if( keylen <= 0 ) cout << "wrong key length: " << argv[ 2 ] << endl, exit( EXIT_FAILURE );
    // проверить - запущена ли сеть qnet?
    int fd = open( "/net", O_RDONLY );
    if( fd < 0 ) {
        // попытаться её запустить
        int rc = system( "mount -Tio-net npm-qnet.so" );
        if( rc == -1 ) cout << "shell could not be run." << endl, exit( EXIT_FAILURE );
        if( WEXITSTATUS( rc ) != 0 ) cout << "there is no net." << endl, exit( EXIT_FAILURE );
        fd = open( "/net", O_RDONLY );
    }
    if( fd < 0 ) cout << "could not use qnet." << endl, exit( EXIT_FAILURE ); 
    close( fd );
    // сосчитать имена хостов в qnet
    DIR* dirp = opendir( "/net" );
    if( dirp == NULL ) cout << "could not use qnet." << endl, exit( EXIT_FAILURE );
    struct dirent* direntp;
    // последовательно по всему списку запустить удалённых агентов
    const char *sAgent = "agent";
    // проверяется наличие агентской программы в каталоге запуска
    if( ( fd = open( sAgent, O_RDONLY ) ) < 0 )
        cout << "agent programm not found." << endl, exit( EXIT_FAILURE ); 
    else close( fd );
    char sPath[ PATH_MAX + 1 ], sComm[ PATH_MAX + 1 ], sHome[ PATH_MAX + 1 ];
    // формирование сетевого имени файла источника
    netmgr_ndtostr( ND2S_LOCAL_STR, ND_LOCAL_NODE, sPath, PATH_MAX + 1 );
    sprintf( sHome, "/net/%s%s/%s", sPath, getcwd( sComm, PATH_MAX + 1 ), argv[ 1 ] );    
    unsigned long long nSumCps = 0;
    // перебираем узлы кластера
    cout << "There are hosts into cluster:" << endl; 
    for( int i = 0; ; i++ ) {
        if( ( direntp = readdir( dirp ) ) == NULL ) break;
        Sutelite *p = new Sutelite( direntp->d_name );
        // проверка существования агента
        sprintf( sPath, "/net/%s/dev/%s", p->host(), sAgent );
        cout << "- " << sPath << " : ";
        if( ( p->fd = open( sPath, O_RDWR ) ) > 0 ) cout << "exist";
        else {
            // запуск ресурс менеджера агента
            sprintf( sComm, "on %s%s %s", 
                        ( netmgr_strtond( p->host(), NULL ) == ND_LOCAL_NODE ? "-n" : "-f" ),
                        p->host(), sAgent );
            system( sComm );
            // проверка успешности запуска агента
            cout << ( ( p->fd = open( sPath, O_RDWR ) ) > 0 ? "open" : "not open !\n" );
        };
        if( p->fd > 0 ) hlist->add( p ); else { delete p; continue; };
        // команды devctl передаётся строкой с завершающим "\0"
        // запрос тактовой частоты хоста
        if( devctl( p->fd, DCMD_CLUST_GTF, sComm, 10, NULL ) != EOK ) perror( "DEVCTL error!" ); 
        cout << "\t-- fr. = " << ( p->cps = atoll( sComm ) ) << endl;
        nSumCps += p->cps;
        // передать файл источник менеджеру хоста
        if( devctl( p->fd, DCMD_CLUST_STS, sHome, strlen( sHome ) + 1, NULL ) < 0 ) perror( "DEVCTL error!" ); 
        // передать длину ключа менеджеру хоста
        sprintf( sComm, "%d", keylen );
        if( devctl( p->fd, DCMD_CLUST_STK, sComm, strlen( sComm ) + 1, NULL ) < 0 ) perror( "DEVCTL error!" ); 
    };
    closedir( dirp );
    // всё, инициализация кластера закончена
    cout << "Number of hosts in cluster = " << hlist->size() << endl;
    // в примере упрощённо длина ключа - не более int64
    unsigned long long nMaxKey = 0, nBegKey = 0, nFinKey = 0;
    for( int i = 0; i < keylen; i++ ) nMaxKey = ( nMaxKey<<8 ) + 0xff;
    pthread_barrier_init( &barrier, NULL, 1 + hlist->size() );
    for( Sutelite *p = hlist->head(); p != NULL; p = hlist->next( p ) ) {
        // вычисляются квоты в зависимости от производительности хостов
        nBegKey = nFinKey + 1;
        nFinKey = nBegKey + (unsigned long long)floor( (double)nMaxKey * p->cps / nSumCps );
        nFinKey = __min( nFinKey, nMaxKey );
        key bKey( keylen ), fKey( keylen );     
        fKey.rshift( nFinKey );
        bKey.rshift( nBegKey );
        // в буфер обмена заносятся 2 ключа: начальный и конечный
        p->buf = new _Uint8t[ keylen * 2 ];
        memcpy( p->buf, (_Uint8t*)bKey, keylen );
        memcpy( p->buf + keylen, (_Uint8t*)fKey, keylen );
        // запуск хоста в работу
        pthread_create( NULL, NULL, &exec, (void*)p );
    };
    // синхронизация завершения всех обрабатывающих хостов
    pthread_barrier_wait( &barrier );
    // вывод результатов
    unsigned long long n = 0;
    for( Sutelite *p = hlist->head(); p != NULL; p = hlist->next( p ) ) {
close( p->fd );
        cout << p->name << ":";
        if( p->nbuf == 0 ) {
            cout << " not found" << endl;
            break;
        }
        _Uint8t *t = p->buf;
        int k = 0;
        for( ; k < p->nbuf / keylen; t += keylen ) {
            key w( keylen, t ); 
            cout << "\t" << w;
            ++k;
            if( k % KEY_OUT_INLINE == 0 ) cout << endl;
        };
        if( k % KEY_OUT_INLINE != 0 ) cout << endl;
        n += k;
    };
    cout << "find " << n << " keys" << endl;
    delete hlist; 
    return EXIT_SUCCESS;
};

